#R library developed and used in Coregistration Simulations
#John Hogland 12/3/2018

library(raster)
library(gstat)

rasterOptions(maxmemory = 1e+09)
rasterOptions(tmpdir = paste(getwd(), "tmp", sep = "/"))
tmpDir(create = TRUE)

###creates spatially correlated raster image. 
#rws = number of rows
#clms = number of colunms
#r = range
#mdl = model-> Sph, Exp, Gau, Mat
#mn = mean values
#vr = variance
#maxDist = maximum distance of spatial correlation (number of cells)
#sim = number of simulations'
getSpCorrImages <- function(rws, clms, r, mdl = "Sph", mn = 0, vr = 1, maxDist = 30, ng = 0, sim = 1) {
    xy <- expand.grid(1:rws, 1:clms)
    names(xy) <- c("x", "y")
    g.dummy <- gstat(formula = z ~ 1, locations = ~x + y, dummy = T, beta = c(mn), model = vgm(psill = vr, model = mdl, range = r, nugget = ng), nmax = maxDist)
    yy <- predict(g.dummy, newdata = xy, nsim = sim)
    outSt <- stack()
    for (k in seq(sim)) {
        yys <- subset(yy, select = c("x", "y", paste("sim", k, sep = "")))
        gridded(yys) = ~x + y
        rs <- raster(yys)
        outSt <- addLayer(outSt, rs)
    }
    return(outSt)
}
#sample a raster and return a spatail points layer
#rs = raster
#sampleSize = number of samples
#type = random or regular'
sampRaster <- function(rs, sampleSize, type = "random") {
    outvl <- NULL
    if (tolower(type) == "random") {
        outvl <- sampleRandom(rs, sampleSize, sp = T)
    }
    else {
        outvl <- sampleRegular(rs, sampleSize, sp = T)
    }
    return(outvl)
}
#shift an existing set of locations to simulate coregistration errors and returns the shifted locations as a dataframe
#pts = Spatial points dataset
#d1 = RMSE in gps expressed as cells
#d2 = RMSE in image rectification expressed as cells'
shiftXY <- function(pts, d1,d2) {
    x <- pts$X
    y <- pts$Y
    cnt <- length(x)
    #d1 <- rnorm(cnt, mean = 0, sd = gpsError)
    #d2 <- rnorm(cnt, mean = 0, sd = imgError)
    b1 <- runif(cnt, min = 0, (2 * pi))
    b2 <- runif(cnt, min = 0, (2 * pi))
    xp1 <- sin(b1) * d1
    yp1 <- cos(b1) * d1
    xp2 <- sin(b2) * d2
    yp2 <- cos(b2) * d2
    xpd <- xp1 + xp2
    ypd <- yp1 + yp2
    df <- data.frame(X = round(xpd + x), Y = round(ypd + y), D = round(sqrt(xpd ** 2 + ypd ** 2)))
    return(df)
}

#get full image matrix
getImgMatrix<-function(rs) {
    return(getValuesBlock(rs,row=1,nrows=nrow(rs),col=1,ncols=ncol(rs),format="matrix",byrow=TRUE))
}

#extract list of point block size matrix
#rcDF = locations
#rs = raster
extractRcLst <- function(rcDf, rsMatrix, blcSize) {
    outLst <- vector(mode = "list", length = nrow(rcDf))
    xc <- rcDf$X
    yc <- rcDf$Y
    for (r in 1:length(outLst)) {
        x <- xc[r]
        y <- yc[r]
        outLst[[r]] <- rsMatrix[y:(y+blcSize),x:(x+blcSize)]#getValuesBlock(rs, row = y, nrows = blcSize, col = x, ncols = blcSize, format = 'matrix')
    }
    return(outLst)
}
#gets multiple values and returns value lst
getMeanBlockValues <- function(lstBlc, lyt, prop) {
    rws <- length(lstBlc)
    mt <- matrix(ncol = 2, nrow = rws)
    for (r in 1:rws) {
        vl <- getMeanBlockValue(lstBlc[[r]], lyt, prop)
        mt[r,] <- vl
    }
    return(mt)
}
#extract values from raster: return dataframe
#rcDf = locations
#rs = raster'
extractRC <- function(rcDf, rs, blcSize, pltLayout = 1, pltProp = 1) {
    rws <- nrow(rcDf)
    mt <- matrix(ncol = 2, nrow = rws)
    xc <- rcDf$X
    yc <- rcDf$Y
    for (r in 1:rws) {
        x <- xc[r]
        y <- yc[r]
        vl <- getMeanBlockValue(getValuesBlock(rs, row = y, nrows = blcSize, col = x, ncols = blcSize, format = 'matrix'), pltLayout, pltProp)
        mt[r,] <- vl
    }
    return(mt)
}
#summarizes values within a block given a specified layout
#blc=matrix of cell values
#lyt = plot layout [1,2,3,4,5,6] 1=center, 2=4 corners, 3=4 random, 4=FIA, 5=9equal,6=corners & center
#prop = proportion of area sampled within the plot extent
getMeanBlockValue <- function(blc, lyt, prop) {
    outVl <- 0
    outProp <- 0
    rws <- nrow(blc)
    clm <- ncol(blc)
    tCells <- sum(!is.na(blc))
    #print(paste("total cells:",tCells,sep=" "))
    if (prop == 1) {
        outVl <- mean(blc, na.rm = TRUE)
        outProp <- 1
    }
    else {
        if (lyt == 1) {
            bSize <- round(sqrt(tCells * prop))
            #print(paste("blockSize:", bSize, sep = " "))
            bSize2 <- bSize - 1
            hb <- round(bSize / 2)
            brw <- round(rws / 2) - hb
            bclm <- round(clm / 2) - hb
            outVl <- mean(blc[brw:(brw + bSize2), bclm:(bclm + bSize2)], na.rm = TRUE)
            ccnt <- sum(!is.na(blc[brw:(brw + bSize2), bclm:(bclm + bSize2)]))
            #print(paste("ccnt:", ccnt, sep = " "))
            outProp <- ccnt / tCells
        }
        else if (lyt == 2) {
            bSize <- round(sqrt((tCells * prop) / 4))
            bSize2 <- bSize - 1
            ccnt <- 0 #bSize**2*4
            #print(paste("BSize2 =",bSize2,sep=" "))
            ms <- blc[1:bSize, 1:bSize]
            m1 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, (clm - bSize2):clm]
            m2 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, 1:bSize]
            m3 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[1:bSize, (clm - bSize2):clm]
            m4 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            outVl <- (m1 + m2 + m3 + m4) / ccnt
            outProp <- ccnt / tCells
        }
        else if (lyt == 3) {
            bSize <- round(sqrt((tCells * prop) / 4))
            bSize2 <- bSize - 1
            ccnt <- 0 #bSize**2*4
            rClm <- as.integer(runif(4, 1, clm - bSize2))
            rRw <- as.integer(runif(4, 1, rws - bSize2))
            sTot <- 0
            for (i in 1:4) {
                r <- rRw[i]
                c <- rClm[i]
                ms <- blc[r:(r + bSize2), c:(c + bSize2)]
                sTot <- sTot + sum(ms, na.rm = TRUE)
                ccnt <- ccnt + sum(!is.na(ms))
                #print(paste("rc:", r, c, sep = " "))
                #print(paste("tot:", sTot, sep = " "))
            }
            outVl <- sTot / ccnt
            outProp <- ccnt / tCells
        }
        else if (lyt == 4) {
            bSize <- round(sqrt((tCells * prop) / 4))
            bSize2 <- bSize - 1
            ccnt <- 0 #bSize**2*4
            hbSize <- round(bSize / 2)
            hc <- round(clm / 2) - hbSize
            hr <- round(rws / 2) - hbSize
            ms <- blc[hr:(hr + bSize2), hc:(hc + bSize2)]
            m1 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, (clm - bSize2):clm]
            m2 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, 1:bSize]
            m3 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[1:bSize, hc:(hc + bSize2)]
            m4 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            sTot <- m1 + m2 + m3 + m4
            #print(paste("tot:", sTot, sep = " "))
            outVl <- sTot / ccnt
            outProp <- ccnt / tCells
        }
        else if (lyt == 5) {
            bSize <- round(sqrt((tCells * prop) / 9))
            bSize2 <- bSize - 1
            ccnt <- 0 #bSize**2*9
            hbSize <- round(bSize / 2)
            hc <- round(rws / 2) - hbSize
            hr <- round(clm / 2) - hbSize
            rClm <- c(1, hc, (clm - bSize))
            rRw <- c(1, hr, (rws - bSize))
            sTot <- 0
            for (i in 1:3) {
                for (j in 1:3) {
                    r <- rRw[i]
                    c <- rClm[j]
                    ms <- blc[r:(r + bSize2), c:(c + bSize2)]
                    sTot <- sTot + sum(ms, na.rm = TRUE)
                    ccnt <- ccnt + sum(!is.na(ms))
                    #print(sTot)
                }
            }
            outVl <- sTot / ccnt
            outProp <- ccnt / tCells
        }
        else if (lyt == 6) {
            bSize <- round(sqrt((tCells * prop) / 5))
            bSize2 <- bSize - 1
            ccnt <- 0 #bSize**2*5
            hbSize <- round(bSize / 2)
            hc <- round(clm / 2) - hbSize
            hr <- round(rws / 2) - hbSize
            ms <- blc[hr:(hr + bSize2), hc:(hc + bSize2)]
            m1 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[1:bSize, 1:bSize]
            m2 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, (clm - bSize2):clm]
            m3 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[(rws - bSize2):rws, 1:bSize]
            m4 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            ms <- blc[1:bSize, (clm - bSize2):clm]
            m5 <- sum(ms, na.rm = TRUE)
            ccnt <- ccnt + sum(!is.na(ms))
            sTot <- m1 + m2 + m3 + m4 + m5
            #print(sTot)
            outVl <- sTot / ccnt
            outProp <- ccnt / tCells
        }
        else {
            outVl <- 0
            outProp <- 0
        }
    }
    return(c(outVl, outProp))
}
#get vector of images
#dir = search directory
#ext=extension of images'
getImgPath <- function(imgDir, ext = ".tif") {
    outVec = c()
    imgNames <- list.files(path = imgDir, pattern = ext)
    for (i in imgNames) {
        ext2 <- extension(i)
        if (tolower(ext2) == ext) {
            outPath <- paste(imgDir, i, sep = "\\")
            outVec = c(outVec, outPath)
        }
    }
    return(outVec)
}
#returns max sill (nugget + psill), max (range) in cells
getSampledMaxSillRange <- function(rsPath, sSize = 20, blSize = 200) {
    r <- raster(rsPath)
    outDf <- data.frame(nm = character(), bnd = double(), mn = double(), sill = double(), rng = double(), nug = double(), cnt = integer())
    rndX <- as.integer(runif(sSize, blSize, nrow(r) - blSize))
    rndY <- as.integer(runif(sSize, blSize, ncol(r) - blSize))
    for (b in 1:nbands(r)) {
        print(paste("Working on band", b, sep = " "))
        mVar <- 0
        mRange <- 0
        mMean <- 0
        mNugget <- 0
        rs <- raster(rsPath, b)
        dn <- 0
        for (bl in 1:sSize) {
            mtVl <- getValuesBlock(rs, row = rndX[bl], nrows = blSize, col = rndY[bl], ncols = blSize, format = 'matrix', byrow = T)
            rs2 <- raster(mtVl)
            mMean <- mMean + cellStats(rs2, mean)
            cellSize <- res(rs2)[1]
            frm <- formula(paste(names(rs2), "1", sep = "~"))
            v <- variogram(frm, as(rs2, "SpatialPixelsDataFrame"))
            f <- tryCatch(fit.variogram(v, vgm("Sph")), warning = function(w) return(NA), error = function(e) return(NA))
            if (is.na(f)) {
                print(paste("Error with", names(r), b, bl, sep = " "))
            }
            else {
                mVar <- mVar + f$psill[2]
                mNugget <- mNugget + f$psill[1]
                mRange <- mRange + (f$range[2] / cellSize)
                dn <- dn + 1
            }
        }
        outDf <- rbind(outDf, data.frame(Name = names(r), Band = b, Mean = mMean / dn, Sill = mVar / dn, Rng = mRange / dn, Nugget = mNugget / dn, cnt = dn))
    }
    names(outDf) <- c("Name", "Band", "Mean", "Sill", "Rng", "Nugget", "Count")
    return(outDf)
}
#creates a random set of cells, returns data frame of row and column
getRandomCells <- function(rs, numCells, insideBuffer,moff) {
    w <- ncol(rs)
    h <- nrow(rs)
    rx <- as.integer(runif(numCells, moff, w - (insideBuffer+moff)))
    ry <- as.integer(runif(numCells, moff, h - (insideBuffer + moff)))
    return(data.frame(X = rx, Y = ry))
}
#perform simulations
SimSampling <- function(ImgPaths, sampSize, blockWidth, shiftGps, shiftImg, prop = 1) {
    outDf <- data.frame(NAME = character(), BAND = double(), BLOCKSIZE = double(), LAYOUT = integer(), EPROP = double(), TPROP=double(), TYPE = integer(), RMSE = double(), MAE = double(), INTERCEPT = double(), SLOPE = double(), R2 = double(), N = integer(), MORI = double(), OVERLAP=double())
    for (i in seq(length(ImgPaths))) {
        pt <- ImgPaths[i]
        print(paste("Working on ", pt, sep = ""))
        rs <- raster(pt)
        nm <- names(rs)
        bndCnt <- nbands(rs)
        d1<- rnorm(sampSize, mean = 0, sd = shiftGps)
        d2 <- rnorm(sampSize, mean = 0, sd = shiftImg)
        d3 <- max(d1)+max(d2)+1
        pnts <- getRandomCells(rs, sampSize, max(blockWidth),d3)
        spnts <- shiftXY(pnts, d1, d2)
        for (b in 1:bndCnt) {
            print(paste("  Band ", b, sep = ""))
            rsb <- convertRasterBandToFloat(raster(pt, b))
            mori <- Moran(rsb, matrix(c(0, 1, 0, 1, 0, 1, 0, 1, 0), nrow = 3)) #rooks case
            rsM <- getImgMatrix(rsb)
            for (bl in blockWidth) {
                print(paste("    Block ", bl, sep = ""))
                shiftMt <- extractRcLst(spnts, rsM, bl)
                #shiftVls <- extractRC(spnts, rsb, bl, 1, 1)
                shiftVls <- getMeanBlockValues(shiftMt, 1, 1)
                vlsMt <- extractRcLst(pnts, rsM, bl)
                chv <- prop[1]
                ovlp = 0
                if (chv == 1 || is.null(chv)) {
                    #vls <- extractRC(pnts, rsb, bl, 1, 1)
                    ovlp <- getOverLap(pnts, spnts, bl)
                    vls <- getMeanBlockValues(vlsMt, 1, 1)
                    tDf <- getLmDf(vls, shiftVls)
                    outDf <- rbind(outDf, data.frame(NAME = nm, BAND = b, BLOCKSIZE = bl, LAYOUT = 0, EPROP = 1, TPROP=1,TYPE = 0, RMSE = tDf$RMSE, MAE = tDf$MAE, INTERCEPT = tDf$INTERCEPT, SLOPE = tDf$SLOPE, R2 = tDf$R2, N = tDf$N, MORI = mori, OVERLAP=ovlp))
                }
                else {
                    for (j in seq(1, 6, 1)) {
                        print(paste("      Layout ", j, sep = ""))
                        for (p in prop) {
                            print(paste("        Proportion ", p, sep = ""))
                            vls <- getMeanBlockValues(vlsMt, j, p)
                            mprop <- mean(vls[, 2])
                            shiftVls2 <- getMeanBlockValues(shiftMt, j, p)
                            tDf <- getLmDf(vls, shiftVls)
                            outDf <- rbind(outDf, data.frame(NAME = nm, BAND = b, BLOCKSIZE = bl, LAYOUT = j, EPROP = p, TPROP = mprop, TYPE = 1, RMSE = tDf$RMSE, MAE = tDf$MAE, INTERCEPT = tDf$INTERCEPT, SLOPE = tDf$SLOPE, R2 = tDf$R2, N = tDf$N, MORI = mori, OVERLAP = ovlp))
                            tDf <- getLmDf(vls, shiftVls2)
                            outDf <- rbind(outDf, data.frame(NAME = nm, BAND = b, BLOCKSIZE = bl, LAYOUT = j, EPROP = p, TPROP = mprop, TYPE = 2, RMSE = tDf$RMSE, MAE = tDf$MAE, INTERCEPT = tDf$INTERCEPT, SLOPE = tDf$SLOPE, R2 = tDf$R2, N = tDf$N, MORI = mori, OVERLAP = ovlp))
                        }
                    }
                }
            }
        }
    }
    removeTmpFiles(h = 0)
    return(outDf)
}
cCalculates mean overlap between shifted points given a block size
getOverLap <- function(pnts, spnts, bl) {
    s <- 0
    ta<-bl^2
    for (i in 1:nrow(pnts)) {
        l1 <- pnts[i,1]
        r1 <- l1 + bl
        b1 <- pnts[i,2]
        t1 <- b1 + bl
        l2 <- spnts[i,1]
        r2 <- l2 + bl
        b2 <- spnts[i,2]
        t2 <- b2 + bl
        ext1 <- extent(l1, r1, b1, t1)
        ext2 <- extent(l2, r2, b2, t2)
        int2 <- intersect(ext1, ext2)
        if (!is.null(int2)) {
            x <- xmax(int2) - xmin(int2)
            y <- ymax(int2) - ymin(int2)
            s <- s + x * y
        }
    }
    return(s / (nrow(pnts)*ta))
}
#performs linear model
getLmDf <- function(vls, shiftVls) {
    md <- lm(vls[, 1] ~ shiftVls[, 1])
    mae <- mean(residuals(md))
    rmse <- mean(residuals(md) ** 2) ** 0.5
    coef <- md$coefficients
    int <- coef[1]
    slp <- coef[2]
    r2 <- (summary(md))$r.squared
    n <- nrow(vls)
    return(data.frame(RMSE = rmse, MAE = mae, INTERCEPT = int, SLOPE = slp, R2 = r2, N = n))
}
#printMoransI
printMoransI <- function(imgPaths, m = matrix(c(0, 1, 0, 1, 0, 1, 0, 1, 0), nrow = 3)) {
    for (i in imgPaths) {
        rs <- raster(i)
        bndCnt <- nbands(rs)
        for (j in 1:bndCnt) {
            rsb <- convertRasterBandToFloat(raster(i, j))
            mori <- Moran(rsb)
            print(paste(i, j, mori, sep = " "))
        }
    }
    removeTmpFiles(h = 0)
}
#setFloat
convertRasterBandToFloat <- function(rsb) {
    dt <- dataType(rsb)
    outRs <- rsb
    if (dt == "FLT4S" || dt == "FLT8S") {
    }
    else {
        outRs <- rsb * 1.0
    }
    return(outRs)
}
#compare to a set number and return a vector of numbers
compareValues <- function(df, fldIndex = 13, stepValue = 40) {
    if (is.character(fldIndex)) {
        fldIndex<-match(fldIndex,names(df))
    }
    rIndx <- 0
    dr2v <- vector(mode = "numeric", length = nrow(df))
    for (r in 1:nrow(df)) {
        r2 <- df[r, fldIndex]
        chx <- r %% stepValue
        if (chx == 1) {
            rIndx <- rIndx + stepValue
        }
        r2c <- df[rIndx, fldIndex]
        dr2v[r] <- r2c - r2
    }
    return(dr2v)
}
#estimate spatial offsets
estOffset <- function(gpsError=6,imageError=7,iterations=1000,sampSize = 200, domainSize = 1000) {
    ss <- 0
    for (i in seq(iterations)) {
        xyDf <- data.frame(X = runif(sampSize, 0, domainSize), Y = runif(sampSize, 0, domainSize))
        sDf <- shiftXY(xyDf, rnorm(sampSize, 0, imageError), rnorm(sampSize, 0, gpsError))
        ss <- ss + sum(sDf$D)
    }
    return(ss / (sampSize*iterations))
}